from allosaurus.am.allosaurus_torch import AllosaurusTorchModel
from allosaurus.am.utils import *
import json
from argparse import Namespace

def read_am(model_path, inference_config):
    """
    load pretrained acoustic model

    :param model_path: path to the
    :return:
    """

    am_config = Namespace(**json.load(open(str(model_path / 'am_config.json'))))

    assert am_config.model == 'allosaurus', "This project only support allosaurus model"

    model = AllosaurusTorchModel(am_config)

    # load weights
    torch_load(model, str(model_path / 'model.pt'))

    # move to device
    if inference_config.device_id >= 0:
        model = model.cuda(inference_config.device_id)

    return model


import wave
import numpy as np
from pathlib import Path
import resampy


def read_audio(filename, header_only=False, channel=0):
    """
    read_audio will read a raw wav and return an Audio object

    :param header_only: only load header without samples
    """

    if isinstance(filename, Path):
        filename = str(filename)

    wf = wave.open(filename)

    # initialize audio
    audio = Audio()

    # set stream basic info
    channel_number = wf.getnchannels()

    # check the input channel is valid
    assert channel < channel_number

    # set wav header
    audio.set_header(sample_rate=wf.getframerate(), sample_size=wf.getnframes(), channel_number=1,
                     sample_width=wf.getsampwidth())

    # set audio
    if not header_only:
        x = wf.readframes(wf.getnframes())

        assert (channel_number <= 2)

        audio_bytes = np.frombuffer(x, dtype='int16')

        # get the first channel if stereo
        if channel_number == 2:
            audio_bytes = audio_bytes[channel::2]

        audio.samples = audio_bytes

        # when some utils piping to stdout, sample size might not be correct (e.g: lame --decode)
        audio.sample_size = len(audio.samples)

    wf.close()

    return audio


def resample_audio(audio, target_sample_rate):
    """
    resample the audio by the target_sample_rate

    :param audio:
    :param target_sample_rate:
    :return:
    """

    # return the origin audio if sample rate is identical
    if audio.sample_rate == target_sample_rate:
        return audio

    new_samples = resampy.resample(audio.samples, audio.sample_rate, target_sample_rate)

    new_audio = Audio(new_samples, target_sample_rate)

    return new_audio


class Audio:

    def __init__(self, samples=None, sample_rate=8000):
        """
        Audio is the basic data structure used in this package.
        It is used to capture fundamental info about audio files such as frequency and samples.

        :param samples:
        :param sample_rate:
        :param stream_name:
        """

        # default parameters
        if samples is None:
            samples = []

        self.sample_rate = sample_rate
        self.channel_number = 1
        self.sample_width = 2

        # segments
        self.segments = []

        # all samples
        self.set_samples(samples)

    def __str__(self):
        wave_info = "<Audio sample rate: "+str(self.sample_rate)+", samples: "\
                    + str(self.sample_size) + ", second: " + str(self.sample_size/self.sample_rate) + " > "
        return wave_info

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return self.sample_size

    def set_samples(self, samples):
        self.samples = samples
        self.sample_size = len(samples)

    def empty(self):
        return self.samples is None or self.sample_size == 0

    def clear(self):
        self.set_samples([])

    def extend(self, new_audio):
        """
        extend wave stream

        :param new_audio:
        :return:
        """

        # resample if sample_rate does not match
        if self.sample_rate != new_audio.sample_rate:
            audio =  resample_audio(new_audio, self.sample_rate)
            samples = audio.samples

        else:
            samples = new_audio.samples

        # extend
        new_samples = np.append(self.samples, samples)
        self.set_samples(new_samples)


    def set_header(self, sample_rate=8000, sample_size=0, channel_number=1, sample_width=2):
        self.sample_rate = sample_rate
        self.sample_size = sample_size
        self.channel_number = channel_number
        self.sample_width = sample_width

    def duration(self):
        return self.sample_size/self.sample_rate


import torch
import torch.nn as nn

class AllosaurusTorchModel(nn.Module):

    def __init__(self, config):
        super(AllosaurusTorchModel, self).__init__()

        self.hidden_size = config.hidden_size
        self.layer_size = config.layer_size
        self.proj_size = config.proj_size

        # decide input feature size
        if config.feat_size == -1:
            corpus_feat_size_dict = list(config.feat_size_dict.values())[0]
            self.feat_size = list(corpus_feat_size_dict.values())[0]

        else:
            self.feat_size = config.feat_size

        assert hasattr(config, 'lang_size_dict'), " config should has the lang_size_dict property"

        self.lang_size_dict = config.lang_size_dict
        self.lang_output_size = dict()

        self.phone_size = config.phone_size

        self.config = config

        self.blstm_layer = nn.LSTM(self.feat_size, self.hidden_size, num_layers=self.layer_size, bidirectional=True)

        self.phone_layer = nn.Linear(self.hidden_size*2, self.phone_size)

        self.phone_tensor = None

    @staticmethod
    def add_args(parser):
        parser.add_argument('--feat_size',   type=int, default=-1, help='input size in the blstm model. if -1, then it is determined automatically by loader')
        parser.add_argument('--hidden_size', type=int, default=320, help='hidden size in the blstm model')
        parser.add_argument('--lang_size',   type=int, default=-1, help='output size in the blstm model, if -1, then it is determined automatically by loader')
        parser.add_argument('--proj_size',   type=int, default=0, help='projection')
        parser.add_argument('--layer_size',  type=int, default=5, help='layer size in the blstm model')
        parser.add_argument('--l2', type=float, default=0.0, help='regularization')
        parser.add_argument('--loss', type=str, default='ctc', help='ctc/warp_ctc/e2e')
        parser.add_argument('--debug_model', type=str, default=False, help='print tensor info for debugging')


    def forward(self, input_tensor, input_lengths, meta=None):
        """

        :param input: an Tensor with shape (B,T,H)
        :lengths: a list of length of input_tensor, if None then no padding
        :meta: dictionary containing meta information (should contain lang_id in this case
        :return:
        """

        #if utt_ids:
            #print("utt_ids {} \n target_tensor {}".format(' '.join(utt_ids), target_tensor))
            #print("input_lengths {}".format(str(input_lengths)))
            #print("target_tensor {}".format(target_tensor))
            #print("target_lengths {}".format(target_lengths))


        # (B,T,H) -> (T,B,H)
        input_tensor = input_tensor.transpose(0, 1)

        # extract lengths
        if input_lengths is None:
            input_lengths = torch.LongTensor([input_tensor.shape[0]]*input_tensor.shape[1])

        # keep the max length for padding
        total_length = input_tensor.size(0)

        #if self.config.loss == 'warp_ctc':
        #target_tensor = torch.cat([target_tensor[idx,:index] for idx, index in enumerate(target_lengths)])

        #if lengths.dim() == 2:
        #    lengths = lengths.squeeze()

        # build each layer

        # (T,B,H) -> PackSequence
        pack_sequence = nn.utils.rnn.pack_padded_sequence(input_tensor, input_lengths)

        # PackSequence -> (PackSequence, States)
        self.blstm_layer.flatten_parameters()

        hidden_pack_sequence, _ = self.blstm_layer(pack_sequence)

        # PackSequence -> (T,B,2H), lengths
        output_tensor, _ = nn.utils.rnn.pad_packed_sequence(hidden_pack_sequence, total_length=total_length)

        # (T,B,2H) -> (T,B,P)
        phone_tensor = self.phone_layer(output_tensor)

        # return (B,T,H) for gathering
        return phone_tensor.transpose(0,1)


import numpy as np
import resampy
from allosaurus.audio import Audio


def feature_cmvn(feature):

    frame_cnt = feature.shape[0]
    spk_sum = np.sum(feature, axis=0)
    spk_mean = spk_sum / frame_cnt
    spk_square_sum = np.sum(feature*feature, axis=0)
    spk_std = (spk_square_sum / frame_cnt - spk_mean * spk_mean) ** 0.5

    return (feature - spk_mean)/spk_std


def feature_window(feature, window_size=3):

    assert window_size == 3, "only window size 3 is supported"

    feature = np.concatenate((np.roll(feature, 1, axis=0), feature, np.roll(feature, -1, axis=0)), axis=1)
    feature = feature[::3, ]

    return feature

def resample_audio(audio, target_sample_rate):
    """
    resample the audio by the target_sample_rate

    :param audio:
    :param target_sample_rate:
    :return:
    """

    # return the origin audio if sample rate is identical
    if audio.sample_rate == target_sample_rate:
        return audio

    new_samples = resampy.resample(audio.samples, audio.sample_rate, target_sample_rate)

    new_audio = Audio(new_samples, target_sample_rate)

    return new_audio


import argparse

def create_allosaurus_parser():

    parser = argparse.ArgumentParser('Phone recognizer')
    parser.add_argument('--device_id', type=int, default=-1, help='specify cuda device id to use, -1 means no cuda and will use cpu for inference')
    parser.add_argument('--model', type=str, default='latest', help='specify which model to use. default is to use the latest model')
    parser.add_argument('--lang_id', type=str, default='ipa',help='specify which language inventory to use for recognition. default is to use all phone inventory')

    return parser

def create_allosaurus_config():
    parser = create_allosaurus_parser()
    args = parser.parse_args()

    return args


from allosaurus.pm.feature import mfcc
from allosaurus.pm.utils import *
import numpy as np

class MFCC:

    def __init__(self, config):

        self.model = config.model

        # feature model config
        self.config = config

        # sample rate
        self.sample_rate = config.sample_rate

        # cmvn
        self.cmvn = config.cmvn

        # samples in each window
        self.window_size =  int(config.window_size * config.sample_rate)

        # overlap between windows
        self.window_shift = int(config.window_shift * config.sample_rate)

        # feature window
        self.feature_window = config.feature_window

        # last complete window starting sample (index of sample)
        self.prev_window_sample = 0

        # last complete mfcc window index (index of window)
        self.prev_window_index = 0

        # list of mfcc features
        self.mfcc_windows = []

        # float32 or float64
        self.dtype = config.dtype

    def __str__(self):
        return "MFCC ("+str(vars(self.config))+")"

    def __repr__(self):
        return self.__str__()


    def compute(self, audio):
        """
        compute feature for audio

        :param audio:
        :return: mfcc feature
        """

        # make sample rate consistent
        audio = resample_audio(audio, self.sample_rate)

        # validate sample rate
        assert self.config.sample_rate == audio.sample_rate, " sample rate of audio is "+str(audio.sample_rate)+" , but model is "+str(self.config.sample_rate)

        # get feature and convert into correct type (usually float32)
        feat = mfcc(audio.samples, samplerate=self.config.sample_rate, numcep=self.config.cep_size, nfilt=self.config.bank_size,
                    lowfreq=self.config.low_freq, highfreq=self.config.high_freq, useEnergy=self.config.use_energy, dither=self.config.dither).astype(self.dtype)

        # apply cmvn if specified
        if self.cmvn == 'speaker':
            feat = feature_cmvn(feat)

        # subsampling and windowing
        if self.feature_window == 3:
            feat = feature_window(feat)

        return feat


from allosaurus.lm.inventory import *
from pathlib import Path
from itertools import groupby

class PhoneDecoder:

    def __init__(self, model_path):
        """
        This class is an util for decode both phones and words

        :param model_path:
        """

        # lm model path
        self.model_path = Path(model_path)

        # create inventory
        self.inventory = Inventory(model_path)

        self.unit = self.inventory.unit

    def compute(self, logits, lang_id=None, blank_factor=1.0):
        """
        decode phones from logits

        :param logits: numpy array of logits
        :param blank_factor:
        :return:
        """

        # apply mask if lang_id specified, this is to restrict the output phones to the desired phone subset
        if lang_id and lang_id != 'ipa':

            mask = self.inventory.get_mask(lang_id)

            logits = mask.mask_logits(logits)

            # print("masking ", str(mask_units))
        else:
            mask = self.inventory.unit

        # greedy decoding to find argmax phones
        decoded_seq = []

        for t in range(len(logits)):

            logit = logits[t]
            logit[0] /= blank_factor

            arg_max = np.argmax(logit)
            decoded_seq.append(arg_max)

        ids = [x[0] for x in groupby(decoded_seq)]

        # ignore when x is <blk> (0)
        cleaned_decoded_seq = [x for x in filter(lambda x: x != 0, ids)]

        phones = ' '.join(mask.get_units(cleaned_decoded_seq))
        return phones


from pathlib import Path
from allosaurus.lm.inventory import Inventory

if __name__ == '__main__':

    model_dir = Path(__file__).parent / 'pretrained'

    models = sorted(model_dir.glob('*'))
    if len(models) == 0:
        print("No models are available, you can maually download a model with download command or just run inference to download the latest one automatically")
        exit(0)

    inventory = Inventory(models[0])

    print("Available Languages")
    for lang in inventory.lang_names:
        print('-', lang)


from pathlib import Path

def get_all_models():
    """
    get all local models

    :return:
    """
    model_dir = Path(__file__).parent / 'pretrained'
    models = list(sorted(model_dir.glob('*'), reverse=True))
    return models

def resolve_model_name(model_name='latest'):
    """
    select the model

    :param model_name:
    :return:
    """

    models = get_all_models()

    # get the latest model in local
    if model_name == 'latest':
        return models[0].name

    # match model name
    for model in models:
        if model.name == model_name:
            return model_name

    return "none"


if __name__ == '__main__':

    models = get_all_models()

    if len(models) == 0:
        print("No models are available, you can maually download a model with download command or just run inference to download the latest one automatically")
    else:
        print("Available Models")
        for i, model in enumerate(models):
            if i == 0:
                print('-', model.name, "(default)")
            else:
                print('-', model.name)


from pathlib import Path
import tarfile
from urllib.request import urlopen
import io
import argparse

def download_model(model_name=None):

    if model_name is None:
        model_name = 'latest'

    model_dir = (Path(__file__).parent) / 'pretrained'

    if not (model_dir / model_name).exists():

        try:
            url = 'https://www.pyspeech.com/static/model/recognition/allosaurus/' + model_name + '.tar.gz'
            print("downloading model ", model_name)
            print("from: ", url)
            print("to:   ", str(model_dir))
            print("please wait...")
            resp = urlopen(url)
            compressed_files = io.BytesIO(resp.read())
            files = tarfile.open(fileobj=compressed_files)
            files.extractall(str(model_dir))

        except:
            print("Error: could not download the model")
            (model_dir / model_name).rmdir()


if __name__ == '__main__':

    parser = argparse.ArgumentParser('a utility to download pretrained models')
    parser.add_argument('-m', '--model', default='latest',  help='specify which model to download. A list of downloadable models are available on Github')

    args = parser.parse_args()

    download_model(args.model)


import json
from allosaurus.lm.mask import *

class Inventory:

    def __init__(self, model_path):

        self.model_path = model_path

        self.lang_names = []

        self.lang2phonefile = dict()

        # load all available inventories
        langs = json.load(open(self.model_path / 'inventory' / 'index.json', 'r', encoding='utf-8'))

        # load all phones list
        self.unit = read_unit(str(self.model_path / 'phone.txt'))

        for lang in langs:

            lang_name = lang['LanguageName']
            phone_text = lang['phonelists']
            self.lang_names.append(lang_name)

            self.lang2phonefile[lang_name.lower()] = phone_text


    def get_mask(self, lang_id=None, approximation=False):

        assert lang_id in self.lang2phonefile, "Language "+lang_id+" is not available !"

        unit_file = self.model_path / 'inventory' / self.lang2phonefile[lang_id]

        target_unit = read_unit(str(unit_file))

        return UnitMask(self.unit, target_unit, approximation)


from allosaurus.app import read_recognizer
from allosaurus.list_model import get_all_models, resolve_model_name
from allosaurus.download import download_model
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser('Allosaurus phone recognizer')
    parser.add_argument('-d', '--device_id', type=int, default=-1, help='specify cuda device id to use, -1 means no cuda and will use cpu for inference')
    parser.add_argument('-m', '--model', type=str, default='latest', help='specify which model to use. default is to use the latest local model')
    parser.add_argument('-l', '--lang', type=str, default='ipa',help='specify which language inventory to use for recognition. default is to use all phone inventory')
    parser.add_argument('-i', '--input', type=str, help='specify your input wav file')

    args = parser.parse_args()

    # check file format
    assert args.input.endswith('.wav'), " Error: Please use a wav file. other audio files can be converted to wav by sox"

    # download specified model automatically if no model exists
    if len(get_all_models()) == 0:
        download_model('latest')

    # resolve model's name
    model_name = resolve_model_name(args.model)
    if model_name == "none":
        print("Model ", model_name, " does not exist. Please download this model or use an existing model in list_model")
        exit(0)

    args.model = model_name

    # create recognizer
    recognizer = read_recognizer(args)

    # run inference
    phones = recognizer.recognize(args.input, args.lang)

    print(phones)


from allosaurus.lm.decoder import PhoneDecoder
import json
from argparse import Namespace

def read_lm(model_path, inference_config):
    """
    read language model (phone inventory)

    :param pm_config:
    :return:
    """

    lm_config = Namespace(**json.load(open(str(model_path / 'lm_config.json'))))

    assert lm_config.model   == 'phone_ipa', 'only phone_ipa model is supported for allosaurus now'
    assert lm_config.backend == 'numpy', 'only numpy backend is supported for allosaurus now'

    model = PhoneDecoder(model_path)
    return model